/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * --------------------------------------------------------------------------*/
#include <stdio.h>

/* rdm.h is required in all files using the RDM Database */
#include "rdmsql.h"

/*! \file perf09.c
 * \brief Source code for the RDM Perf09 RDMSQL example
 */
 
/*!
 * \mainpage Perf09_RDMSQL
 * \author Daigoro F. Toyama, Raima Sr. Engineer
 * \date 2012
 *
 * \section Description
 * EXAMPLE - Perf09 RDMSQL
 *
 * The Perf09 performance example creates 50,000 rows in a simple table
 * in an on-disk database.  It then spawns three threads.  The first one
 * updates rows in the table inside transactions with 1000 updates per
 * transaction.  The second thread attempts to read all the records to 
 * sum their contents.  The third thread just periodically reports the
 * results from the other two threads.  The reading thread uses standard
 * locks.  This can be compared against the performance in Perf10 where
 * the reading thread uses Read-Only Transactions.
 *
 * \section Perf09_db Database
 * The DDL (Database Definition Language) specification for the database used
 * in this example is located in the create_database() function in this file.
 * The example utilizes the SQL DDL syntax and specifies an on-disk database
 * with a single table (simple).  The table has a single column (int_col).
 * There are no index definitions for this database.
 *
 * \section RDM SQL Functions Used
 * The following RDM functions are illustrated:
 *   * rsqlTFSInitEx/rsqlTFSTerm
 *   * rsqlAllocConn/rsqlFreeConn
 *   * rsqlOpenDB/rsqlCloseDB
 *   * rsqlInitDB
 *   * rsqlAllocStmt/rsqlFreeStmt
 *   * rsqlPrepare
 *   * rsqlBindParam
 *   * rsqlBindNamedParam
 *   * rsqlExecute
 *   * rsqlExecDirect
 *   * rsqlFetch
 *   * rsqlTransStart
 *   * rsqlTransCommit
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 */
#define NumRows 50000

static const int32_t transactionSize = 1000;

/*! \brief Create a test database
 *
 */
static RSQL_ERRCODE create_database(void)
{
    RSQL_ERRCODE rsqlCode;
    HCONN        hDbc;
    HSTMT        hStmt;
    int16_t      ii;

    const char *cmds[] = {
        "CREATE DATABASE perf09_db",
        "CREATE TABLE simple (id INTEGER PRIMARY KEY, int_col INTEGER)",
        "COMMIT"
    };

    (void) rsqlAllocConn(&hDbc);
    (void) rsqlAllocStmt(hDbc, &hStmt);

    /* Attempt to drop the database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    (void) rsqlExecDirect(hStmt, "DROP DATABASE perf09_db");
    (void) rsqlExecDirect(hStmt, "COMMIT");

    for (ii = 0; ii < sizeof(cmds) / sizeof(cmds[0]); ii++)
    {
        rsqlCode = rsqlExecDirect(hStmt, cmds[ii]);
        if (rsqlCode != errSUCCESS)
            break;
    }

    /* Close the newly created database */
    (void) rsqlCloseDB(hDbc, "perf09_db");

    (void) rsqlFreeStmt(hStmt);
    (void) rsqlFreeConn(hDbc);

    return errSUCCESS;
}

/*! \brief Insert 50,000 simple records into the perf09_db database
 *
 * This function adds 50,000 simple records into the on-disk perf09 database.  The
 * simple table contains only a single integer column with no index.  The function
 * will calculate and display the total time elapsed while inserting the records
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
static RSQL_ERRCODE add_records(
    HCONN   hDbc)       /**< [in] RDM SQL Connection handle */
{
    HSTMT         hStmt;
    int32_t       id;
    int32_t       ii;
    RSQL_ERRCODE  rsqlCode = errSUCCESS;

    /* allocates insert statement */
    (void) rsqlAllocStmt(hDbc, &hStmt);

    /* Prepare insert statement */
    (void) rsqlPrepare(hStmt,  "INSERT INTO simple VALUES ?, 0");
    (void) rsqlBindParam(hStmt, 1, tINTEGER, &id, 0);

    /* Loop to add 50,000 new simple record */
    for(ii = 0; ii < NumRows; ii++)
    {
        /* Set values for the new record */
        id = ii;

        /* Add the record to the database */
        rsqlCode = rsqlExecute(hStmt);
        if (rsqlCode != errSUCCESS)
        {
            printf("Error has occurred exeucting the insert statement");
            break;
        }
    }

    (void) rsqlTransCommit(hDbc);
    (void) rsqlFreeStmt(hStmt);

    return rsqlCode;
}

/*! \brief Initialize the RDM SQL library for use in the perf09 example
 *
 * This function initializes the RDM Transactional File Server (TFS) to use 
 * the single-user transactional DIRECT-LINK implementation.  It also creates
 * an RDM runtime task and opens the "perf09_db" database in exclusive ('x') mode.
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
static RSQL_ERRCODE initialize(
    HCONN    *hDBC)   /**< [out] Pointer to the connection handle to be initialized */
{
    RSQL_ERRCODE rsqlCode;

    *hDBC = NULL;

    /* Initialize the TFS to use the DIREC-LINK configuration (TFST) */    
    (void) rsqlTFSInitEx(NULL, NULL, TFS_TYPE_TFS);
    rsqlCode = create_database();
    if(rsqlCode != errSUCCESS)
        return rsqlCode;

    (void) rsqlAllocConn(hDBC);
    (void) rsqlOpenDB(*hDBC, "perf09_db", "x");

    (void) add_records(*hDBC);

    (void) rsqlCloseDB(*hDBC, "perf09_db");

    /* Display information */
    printf("Initializing with DIREC-LINK configuration\n");
    printf("\tOne, simple record type.\n");
    printf("\tOn disk\n");
    printf("\tOpened in shared mode\n");
    printf("\t\t\tHit <Enter> to exit the example\n");

    return rsqlCode;
}

/*! \brief Cleanup the RDM runtime library
 *
 * This functions closes all open databases and cleans up the RDM connection
 * handle used in the perf09 example.  It also terminates the DIRECT-LINK TFS
 * used in the example. 
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
static void cleanup(
    HCONN hDBC)   /**< [in] Connection handle to be closed/cleaned up */

{
    if (hDBC)
    {
        /* Close any open databases */
        rsqlCloseDB(hDBC, "perf09_db");

        /* Free the connection handle */
        rsqlFreeConn(hDBC);
    }

    /* Terminate the DIRECT-LINK TFS */
    rsqlTFSTerm();
}

static int16_t quit = 0;
static int64_t rowsUpdated = 0;
static int64_t queriesExecuted = 0;

/* Updater thread */
static void *updater_thread(
    void *ctx)
{
    HCONN   hDbc;
    HSTMT   hStmt;
    int32_t minid_parm;
    int32_t minid;

    printf("Updating rows in transactions of size %d...\n", transactionSize);

    /* Create a connection for this thread */
    (void) rsqlAllocConn(&hDbc);
    (void) rsqlOpenDB(hDbc, "perf09_db", "s");

    (void) rsqlAllocStmt(hDbc, &hStmt);

    (void) rsqlPrepare(hStmt, "UPDATE simple SET int_col = int_col + 1 WHERE id >= :minid AND id < :minid + 1000");
    (void) rsqlBindNamedParam(hStmt, "minid", tINTEGER, &minid_parm, NULL, NULL);

    minid = 0;

    while (!quit)
    {
        if (minid >= NumRows)
            minid = 0;

        (void) rsqlTransStart(hDbc, "updaterTrans");

        minid_parm = minid;
        (void) rsqlExecute(hStmt);

        (void) rsqlTransCommit(hDbc);

        rowsUpdated += transactionSize;
        minid += transactionSize;
    }

    printf("Ending updates\n");

    (void) rsqlFreeStmt(hStmt);

    (void) rsqlCloseDB(hDbc, "perf09_db");
    (void) rsqlFreeConn(hDbc);

    return NULL;
}

/*! \brief Thread function for reading all the records to sum their contents
 *
 * The function reads all the records to sum thier contents.
 */
static void *reader_thread(
    void *ctx)
{
    HCONN    hDbc;
    HSTMT    hStmt;
    char     commandText[80];

    /* Create a connection for this thread */
    (void) rsqlAllocConn(&hDbc);
    (void) rsqlOpenDB(hDbc, "perf09_db", "s");

    (void) rsqlAllocStmt(hDbc, &hStmt);

    sprintf(commandText, "SELECT SUM(int_col) FROM Simple");

    printf("Reading the contents of the table...\n");

    while (!quit)
    {
        (void) rsqlExecDirect(hStmt, commandText);
        queriesExecuted++;
    }

    printf("Ending reading\n");

    (void) rsqlFreeStmt(hStmt);

    (void) rsqlCloseDB(hDbc, "perf09_db");
    (void) rsqlFreeConn(hDbc);

    return NULL;
}

/*! \brief Thread function for reporting the results
 *
 * The function periodically reports results from the other two threads.
 */
static void *reporter_thread(
    void *ctx)
{
    int32_t secs;

    for (secs = 1; ; secs++)
    {
        sleep(1);

        if (quit)
            break;

        printf("Updated %ld rows (%ld/sec) and executed %ld queries (%.2lf/sec) in %d secs\n", rowsUpdated, (rowsUpdated / secs), queriesExecuted, ((double) queriesExecuted / secs), secs);
    }

    return NULL;
}

/*! \brief Main function for perf09 example
 *
 * The function initializes the RDM SQL environment and runs the create, read,
 * update, and delete operations.
 *
 * @return Returns 0 on success, and 1 if there were any errors
 */
int32_t main(void)
{
    RSQL_ERRCODE   rsqlCode;
    HCONN          hDBC;
    pthread_t      updater_tid;
    pthread_t      reader_tid;
    pthread_t      reporter_tid;

    /* Display the example title */
    printf("%s", RDM_EXAMPLE_DESC("Perf09"));

    /* Initialize the TFS, task, and open/initialize the database */
    rsqlCode = initialize(&hDBC);
    if (rsqlCode != errSUCCESS)
    {
        fprintf(stderr, "Application is aborting due to initialization error\n");
        cleanup(hDBC);
        return EXIT_FAILURE;
    }

    (void) pthread_create(&updater_tid, NULL, &updater_thread, NULL);
    (void) pthread_create(&reader_tid, NULL, &reader_thread, NULL);
    (void) pthread_create(&reporter_tid, NULL, &reporter_thread, NULL); 

    /* Wait for user to hit <Enter> */
    getchar();

    /* Indicate to the threads that they should shut down */
    quit = 1;

    /* Wait for each thread to finish */
    pthread_join(updater_tid, NULL);
    pthread_join(reader_tid, NULL);
    pthread_join(reporter_tid, NULL);

    cleanup(hDBC);

    printf("\nPress Enter to terminate\n");
    getchar();

    return EXIT_SUCCESS;
}
